/* "shuffle.c" copyright 1998 daniel risacher  */


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/file.h>
#include <sys/types.h>
#include <signal.h>
#include <errno.h>
#include <time.h>
#include <pwd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <string.h>

#include "config.h"
#include "common.h"

#include "commands.h"
#include "hardware.h"
#include "info.h"
#include "shuffle.h"
#include "cdtool.h"
#include "util.h"

/* globals for this module */
static int   iDone      = FALSE;
static int   iNext      = FALSE;
static int   iBEVerbose = FALSE;
static char  *pszName   = NULL;

/************************************************************************/
/* Procedure:  huphdl
 * Purpose:    handler for SIGHUP
 * 
 * Inputs:     
 * Outputs:    TRUE to iNext flag -- do next track
 * Returns:    
 * Notes:  
 *   1.  Sending a HANGUP (SIGHUP,1) will cause cdshuffle to skip
 *       to the next track.  This may be used for debug or just if
 *       you don't like a song.  
 */
/************************************************************************/
void huphdl ()
{
    signal (SIGHUP, huphdl);
    iNext = TRUE;
    if (iBEVerbose)
	fprintf (stderr, "%s:  Hangup received....\n", pszName);
}

/************************************************************************/
/* Procedure:  inthdl
 * Purpose:    handler for SIGINT
 * 
 * Inputs:     
 * Outputs:    TRUE to iDone flag -- finished with processing
 * Returns:    
 * Notes:  
 *   1.  Sending an INTERRUPT (SIGINT,2) will cause cdshuffle to 
 *       stop playing, remove its PID file, and exit. 
 */
/************************************************************************/
void inthdl ()
{
    signal (SIGINT, inthdl);
    iDone = TRUE;
    if (iBEVerbose)
	fprintf (stderr, "%s:  Interrupt received....\n", pszName);

}

/************************************************************************/
/* Procedure:  wait_track
 * Purpose:    to wait for a track to finish
 * 
 * Inputs:     program name, CD file des., track number
 * Outputs:    none
 * Returns:    pointer to temp string
 * Notes:  
 *   1.  Polls the CD-ROM at a 1 second interval waiting for a 
 *       track to finish....
 */
/************************************************************************/
int wait_track(char* progname, int cdfd, int t)
{
  struct cdrom_subchnl subchnl;
  DEBUGMSG("%s:  waiting on track %d", progname, t);

  while (1) {
    
    subchnl.cdsc_format = CDROM_MSF;
    if (ioctl(cdfd, CDROMSUBCHNL, &subchnl) == -1) {
	if (errno == EINTR)
	    return TRUE;

	ERRORMSG("%s: ioctl cdromsubchnl: %s", progname, strerror(errno));
	return (-1);
    }
    
    switch ( subchnl.cdsc_audiostatus ) {
    case CDROM_AUDIO_PAUSED:
    case CDROM_AUDIO_PLAY:
      sleep(1);
      if (iDone == TRUE) {
	  DEBUGMSG("%s:  done with audio r=0x%X", 
            progname, subchnl.cdsc_audiostatus);
	  return TRUE;
      }

      if (iNext == TRUE) {
	  DEBUGMSG("%s:  next track processing r=0x%X",
	    progname, subchnl.cdsc_audiostatus);
	  iNext = FALSE;
	  return TRUE;
      }
      break;

    case CDROM_AUDIO_COMPLETED:  
	DEBUGMSG("%s:  Audio Completed r=0x%X", 
	  progname, subchnl.cdsc_audiostatus);
	return TRUE;

    default:
	DEBUGMSG("%s:  default r=0x%X", progname, subchnl.cdsc_audiostatus);
        return TRUE;
    }
  }
}

/************************************************************************/
/* Procedure:  do_shuffle
 * Purpose:    to play a CD-ROM in shuffle mode
 * 
 * Inputs:     program name, CD file descriptor, max tracks to play,
 *             count of tracks, verbose flag (T/F), , CD device,
 *             CR/CRLF flag
 * Outputs:    none
 * Returns:    pointer to temp string
 * Notes:  
 *   1.  Calls wait_track to wait for a track to finish playing....
 *   2.  This function should be modified for the following:
 *       A) Additon of a count of times to shuffle and max tracks to play.
 *       B) Change to call read_hw() as read_hw() includes additional 
 *          CD-ROM checks (e.g., if mounted).
 */
/************************************************************************/
void do_shuffle(cdtool_t *cdt, int cdfd, int iCDNum, int iTracks, int iCount)
{
    /* struct cdrom_subchnl subchnl; */
  struct cdrom_tochdr tochdr;
  int tracks[100];
  int nent_tracks;
  int i, j, tmp, r;
  cdhw_t *hw;
  int iSaveTracks;
  int status;

  pid_t my_pid;
  FILE  *fpPID;
  char  caBuf[255];
  struct passwd	*pw;
  char	*homedir;

  iSaveTracks = iTracks;
  pszName = cdt->progname;
  iBEVerbose = cdtool_show_verbose;
  status = 0;

  pw=getpwuid(getuid());
  if (!pw)
  {
      fprintf(stderr, "Username not in passwd database - aborting\n");
      exit(1);
  }
  homedir=malloc(strlen(pw->pw_dir)+1);
  if (!homedir)
  {
      fprintf(stderr, "Error allocating memory\n");
      exit(1);
  }
  strcpy(homedir, pw->pw_dir);

  DEBUGMSG("%s.do_shuffle:  file=%d, num=%d, trk=%d, cnt=%d, V=%d",
    cdt->progname, cdfd, iCDNum, iTracks, iCount, cdtool_show_verbose);

  hw = check_disc (cdt, cdfd, &i);
  if (i == TRUE) {
        
      /* get number of tracks, if not available, error... */
      if ( ioctl(cdfd, CDROMREADTOCHDR, &tochdr) == -1 ) {
	  fprintf(stderr, "%s: ioctl cdromreadtochdr\n", cdt->progname);
	  return;
      }
      
      nent_tracks = tochdr.cdth_trk1;

      /* update the pid file "PID\nTIME\nFILE\n" */
      signal (SIGINT, inthdl);
      signal (SIGHUP, huphdl);
      my_pid = getpid();
      snprintf (caBuf, sizeof(caBuf)-1, "%s%s%d", homedir, PID_FILE, iCDNum);
      fpPID = fopen (caBuf, "w");
      if (fpPID == NULL)
	  ERRORMSG("%s:  Error opening pid file %s: %s", cdt->progname,
            caBuf, strerror(errno));
      else {
	  if (iBEVerbose == TRUE)
	      fprintf(stderr, "%s:  Writing PID file %s\n",
                      cdt->progname, caBuf);
	  sprintf (caBuf, "%ld\n", (unsigned long) my_pid);
	  fputs (caBuf, fpPID);
	  fclose (fpPID);
      }

      /* repeat play N time (if N = -1, do forever) */
      iDone = FALSE;
      while (iDone == FALSE) {

          DEBUGMSG("do_shuf:  tracks=%d repeat=%d verbose=%d",
	    iTracks, iCount, cdtool_show_verbose);

	  /* generate random list of tracks to play */
	  for (i=0; i<nent_tracks; i++)
	      tracks[i] = i+1;
	  
	  srand(time(NULL));
	  for (j=0; j<2; j++) {
	      for (i=0; i<nent_tracks; i++) {
		  r = (int) (nent_tracks * (rand() / (RAND_MAX+1.0)));

		  DEBUGMSG("do_shuf:  j=%d, i=%d, r=%d, tracks=%d",
                    j, i, r, nent_tracks);

		  tmp = tracks[j];
		  tracks[j] = tracks[r];
		  tracks[r] = tmp;
	      }
	  }

	  for (i=0; i<nent_tracks; i++) {
	      DEBUGMSG("do_shuf:  I=%02d track=%02d", i, tracks[i]);
	  }
	  
	  /* play each track in random order */
	  for (i=0; i<nent_tracks; i++) {
	      DEBUGMSG("do_shuf:  playing %d", tracks[i]);
	      if (cdtool_show_verbose)
		  INFOMSG("%s playing track %d", cdt->progname, tracks[i]);
	      
	      /* this plays only this track */
              if (!do_play(cdt, cdfd, tracks[i], tracks[i], NULL)) {
                  /* error */
                  iDone = TRUE;
                  break;
              }

	      /* if we are doing 1 track only, exit */
	      if ((iCount <= 1) && (iTracks == 1) &&
		  (iSaveTracks == 1))
		  break;

	      wait_track(cdt->progname, cdfd, tracks[i]);

	      /* if we have played the desired # of tracks, DONE */
	      iTracks--;
	      if (iTracks <= 0) {
		  iTracks = iSaveTracks;
		  break;
	      }

	      if (iDone == TRUE)
		  break;
	  }
	  
	  /* determine if we are to repeat the disc, iCount==0 is forever */
	  if (iCount > 0) {
	      iCount --;
	      if (iCount == 0)
		  break;
	  }
      }
  }
  free_hw_buf(hw);
  
  if ( ioctl(cdfd, CDROMSTOP) == -1 ) {
      fprintf(stderr, "%s: ioctl cdromstop\n", cdt->progname);
  }

  snprintf (caBuf, sizeof(caBuf)-1, "%s%s%d", homedir, PID_FILE, iCDNum);
  if (iBEVerbose)
      fprintf (stderr, "%s:  removing lock file %s\n", cdt->progname, caBuf);
  unlink (caBuf);

}
